const std = @import("std");

const maxMemory = 0x8000;
const regSize = 8;
const maxStackSize= 32;
const regStart = 32768;
const regEnd = 32775;

const VmState = enum {
    Reset,
    Running,
    Paused,
    Halted,
    Error,
    AwaitingInput,
};

const VmError = error {
    PoppedEmptyStack,
    PushedMaxStack,
    InvalidAddress,
    InvalidRegValue,
    Unknown,
};

pub const Vm = struct {
    mem: [maxMemory]u16,
    reg: [regSize]u16,
    pc: u16,

    stack: [maxStackSize]u16,
    stackSize: usize,

    inputBuffer: std.ArrayList(u8),

    allocator: std.mem.Allocator,
    state: VmState,

    const outWriter = std.io.getStdOut().writer();
    const inReader = std.io.getStdIn().reader();

    fn pop(self: *Vm) VmError!u16 {
        if (self.stackSize == 0) {
            return VmError.PoppedEmptyStack;
        }
        const v = self.stack[self.stackSize - 1];
        self.stackSize -= 1;
        return v;
    }

    fn push(self: *Vm, v: u16) !void {
        if (self.stackSize == maxStackSize) {
            return VmError.PushedMaxStack;
        }
        self.stack[self.stackSize] = v;
        self.stackSize += 1;
    }

    fn getVal(self: *Vm, o: u16) !u16 {
        const val = self.mem[self.pc + o];
        if (val < regStart) {
            return val;
        } else if (val <= regEnd) {
            return self.reg[val - regStart];
        }
        return VmError.InvalidAddress;
    }

    fn getReg(self: *Vm, o: u16) !*u16 {
        const val = self.mem[self.pc + o];
        if (val < regStart or val > regEnd) {
            return VmError.InvalidRegValue;
        }
        return &self.reg[val - regStart];
    }

    pub fn run(self: *Vm) !void {
        self.state = VmState.Running;
        while (self.state == VmState.Running) {
            switch (self.mem[self.pc]) {
            // halt: 0
            0 => self.state = VmState.Halted,
            // set: 1 a b
            1 => {
                const r = try self.getReg(1);
                r.* = try self.getVal(2);
                self.pc += 3;
            },
            // push: 2 a
            2 => {
                try self.push(try self.getVal(1));
                self.pc += 2;
            },
            // pop: 3 a
            3 => {
                const r = try self.getReg(1);
                r.* = try self.pop();
                self.pc += 2;
            },
            // eq: 4 a b c
            4 => {
                const r = try self.getReg(1);
                r.* = if (try self.getVal(2) == try self.getVal(3)) 1 else 0;
                self.pc += 4;
            },
            // gt: 5 a b c
            5 => {
                const r = try self.getReg(1);
                r.* = if (try self.getVal(2) > try self.getVal(3)) 1 else 0;
                self.pc += 4;
            },
            // jmp: 6 a
            6 => self.pc = self.mem[self.pc + 1],
            // jt: 7 a b
            7 => self.pc = if (try self.getVal(1) != 0) try self.getVal(2) else self.pc + 3,
            // jf: 8 a b
            8 => self.pc = if (try self.getVal(1) == 0) try self.getVal(2) else self.pc + 3,
            // add: 9 a b c
            9 => {
                const r = try self.getReg(1);
                r.* = (try self.getVal(2) +% try self.getVal(3)) % regStart;
                self.pc += 4;
            },
            // mult: 10 a b c
            10 => {
                const r = try self.getReg(1);
                r.* = (try self.getVal(2) *% try self.getVal(3)) % regStart;
                self.pc += 4;
            },
            // mod: 11 a b c
            11 => {
                const r = try self.getReg(1);
                r.* = (try self.getVal(2) % try self.getVal(3));
                self.pc += 4;
            },
            // and: 12 a b c
            12 => {
                const r = try self.getReg(1);
                r.* = (try self.getVal(2) & try self.getVal(3));
                self.pc += 4;
            },
            // or: 13 a b c
            13 => {
                const r = try self.getReg(1);
                r.* = (try self.getVal(2) | try self.getVal(3));
                self.pc += 4;
            },
            // not: 14 a b
            14 => {
                const r = try self.getReg(1);
                r.* = ~(try self.getVal(2)) & 0x7FFF;
                self.pc += 3;
            },
            // rmem: 15 a b
            15 => {
                const r = try self.getReg(1);
                const b: usize = @intCast(try self.getVal(2));
                r.* = self.mem[b];
                self.pc += 3;
            },
            // wmem: 16 a b
            16 => {
                self.mem[try self.getVal(1)] = try self.getVal(2);
                self.pc += 3;
            },
            // call: 17 a
            17 => {
                try self.push(self.pc + 2);
                self.pc = try self.getVal(1);
            },
            // ret: 18
            18 => self.pc = try self.pop(),
            // out: 19 a
            19 => {
                const a: u8 = @intCast(try self.getVal(1));
                try outWriter.print("{c}", .{a});
                self.pc += 2;
            },
            // in: 20 a
            20 => {
                if (self.inputBuffer.items.len == 0) {
                    self.inputBuffer = std.ArrayList(u8).init(self.allocator);
                    try inReader.streamUntilDelimiter(self.inputBuffer.writer(), '\n', null);
                    try self.inputBuffer.append('\n');
                }
                const r = try self.getReg(1);
                r.* = self.inputBuffer.items[0];
                _ = self.inputBuffer.orderedRemove(0);
                if (self.inputBuffer.items.len == 0) {
                    self.inputBuffer.deinit();
                }
                self.pc += 2;
            },
            // noop: 21
            21 => self.pc += 1,
            else => {
                try outWriter.print("Unknown opcode {d}\n", .{self.mem[self.pc]});
                self.state = VmState.Error;
            }
            }
        }
    }
};

pub fn init() !Vm {
    var vm = Vm {
        .mem = [_]u16{0} ** maxMemory,
        .reg = [_]u16{0} ** regSize,
        .pc = 0,
        .stack = [_]u16{0} ** maxStackSize,
        .stackSize = 0,
        .allocator = std.heap.GeneralPurposeAllocator(.{}).init.backing_allocator,
        .state = VmState.Reset,
        .inputBuffer = undefined,
    };

    const cwd = std.fs.cwd();
    const fp = try cwd.openFile("challenge.bin", .{});
    defer fp.close();

    const sz = (try fp.stat()).size;
    const buf = try fp.readToEndAlloc(vm.allocator, sz);
    defer vm.allocator.free(buf);

    var i: usize = 0;
    var j: usize = 0;
    while (i < sz) : (i += 2) {
        const a: u16 = @intCast(buf[i]);
        const b: u16 = @intCast(buf[i+1]);
        vm.mem[j] = a | (b << 8);
        j += 1;
    }

    return vm;
}
